'use strict';

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault")["default"];
var _createForOfIteratorHelper2 = _interopRequireDefault(require("@babel/runtime/helpers/createForOfIteratorHelper"));
var _require = require('../store/RelayErrorTrie'),
  RelayFieldError = _require.RelayFieldError;
var RelayFeatureFlags = require('./RelayFeatureFlags');
function handleResolverErrors(environment, relayResolverErrors, throwOnFieldError) {
  var _iterator = (0, _createForOfIteratorHelper2["default"])(relayResolverErrors),
    _step;
  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var resolverError = _step.value;
      environment.relayFieldLogger(resolverError);
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }
  if (RelayFeatureFlags.ENABLE_FIELD_ERROR_HANDLING_THROW_BY_DEFAULT || throwOnFieldError) {
    throw new RelayFieldError("Relay: Unexpected resolver exception", relayResolverErrors.map(function (e) {
      return {
        message: e.error.message
      };
    }));
  }
}
function handleFieldErrors(environment, errorResponseFields, shouldThrow) {
  var _iterator2 = (0, _createForOfIteratorHelper2["default"])(errorResponseFields),
    _step2;
  try {
    for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
      var fieldError = _step2.value;
      environment.relayFieldLogger(fieldError);
    }
  } catch (err) {
    _iterator2.e(err);
  } finally {
    _iterator2.f();
  }
  if (shouldThrow) {
    throw new RelayFieldError("Relay: Unexpected response payload - this object includes an errors property in which you can access the underlying errors", errorResponseFields.map(function (event) {
      switch (event.kind) {
        case 'relay_field_payload.error':
          return event.error;
        case 'missing_expected_data.throw':
          return {
            message: 'Missing expected data'
          };
        case 'missing_expected_data.log':
          return {
            message: 'Missing expected data'
          };
        default:
          event.kind;
          throw new Error('Relay: Unexpected event kind');
      }
    }));
  }
}
function handleMissingRequiredFields(environment, missingRequiredFields) {
  switch (missingRequiredFields.action) {
    case 'THROW':
      {
        var _missingRequiredField = missingRequiredFields.field,
          path = _missingRequiredField.path,
          owner = _missingRequiredField.owner;
        environment.relayFieldLogger({
          kind: 'missing_required_field.throw',
          owner: owner,
          fieldPath: path
        });
        throw new Error("Relay: Missing @required value at path '".concat(path, "' in '").concat(owner, "'."));
      }
    case 'LOG':
      missingRequiredFields.fields.forEach(function (_ref) {
        var path = _ref.path,
          owner = _ref.owner;
        environment.relayFieldLogger({
          kind: 'missing_required_field.log',
          owner: owner,
          fieldPath: path
        });
      });
      break;
    default:
      {
        missingRequiredFields.action;
      }
  }
}
function handlePotentialSnapshotErrors(environment, missingRequiredFields, relayResolverErrors, errorResponseFields, throwOnFieldError) {
  if (relayResolverErrors.length > 0) {
    handleResolverErrors(environment, relayResolverErrors, throwOnFieldError);
  }
  if (missingRequiredFields != null) {
    handleMissingRequiredFields(environment, missingRequiredFields);
  }
  if (errorResponseFields != null) {
    handleFieldErrors(environment, errorResponseFields, throwOnFieldError);
  }
}
module.exports = handlePotentialSnapshotErrors;